!:::::      ,,,,,BANDO.....  -*- f90 -*-
!
!
!
!**********************************************************************
!        1         2         3         4         5         6         7*
!**********************************************************************
!**                                                                 ***
!**                                                                 ***
module bando
!**                                                                 ***
!**                                                                 ***
!**********************************************************************
!**                                                                 ***
!**      program description of BANDO                               ***
!**                                                                 ***
!**      Band optimizes a matrix using the                          ***
!**      Gibbs-Pool-Stockmeyer                                      ***
!**      algorithm.                                                 ***
!**                                                                 ***
!**********************************************************************
!**                                                                 ***
!**    Copyright (C) 1998, 1999 University Karlsruhe                ***
!**    AUTHOR : Detlev Zundel                                       ***
!**                                                                 ***
!**    This file is part of LINSOL.                                 ***
!**    LINSOL is free software; you can redistribute it and/or      ***
!**    modify it under the terms of the GNU General Public License  ***
!**    as published by the Free Software Foundation; either         ***
!**    version 2 of the License, or (at your option) any later      ***
!**    version.                                                     ***
!**                                                                 ***
!**    LINSOL is distributed in the hope that it will be useful,    ***
!**    but WITHOUT ANY WARRANTY; without even the implied warranty  ***
!**    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.      ***
!**    See the GNU General Public License for more details.         ***
!**                                                                 ***
!**    You should have received a copy of the GNU General Public    ***
!**    License along with LINSOL; if not, write to the Free         ***
!**    Software Foundation, Inc., 59 Temple Place - Suite 330,      ***
!**    Boston, MA  02111-1307, USA.                                 ***
!**                                                                 ***
!**********************************************************************

  use constants
  use comm_ifc

  implicit none

  contains

    ! Subroutine : bando
    ! Purpose    : Main subroutine for the band optimizer
    !
    !
    ! List of Formal Parameters :
    !
    !---------I------I-----I------------------------------------------
    ! NAME    I TYPE I I/O I MEANING
    !---------I------I-----I------------------------------------------
    ! graph   I int  I in  I array with indices
    !---------I------I-----I------------------------------------------
    ! vertix  I int  I in  I array with matrix-structure-information
    !---------I------I-----I------------------------------------------
    ! lmatbk  I int  I in  I Array with nr. of matrix lines per proc.
    !---------I------I-----I------------------------------------------
    ! permut  I int  I out I Permutation vector
    !---------I------I-----I------------------------------------------
    ! iverboseI int  I in  I Verbose level
    !---------I------I-----I------------------------------------------
    ! iounit  I int  I in  I Unit for diagnostic messages
    !---------I------I-----I------------------------------------------
    ! myproc  I int  I in  I Logical number of this proc. (1-nproc)
    !---------I------I-----I------------------------------------------
    ! nproc   I int  I in  I Number of total processors
    !---------I------I-----I------------------------------------------
    ! tids    I int  I in  I Array of physical processor IDs
    !---------I------I-----I------------------------------------------
    ! nmsg    I int  I i/o I Monotonously increasing Message nr.
    !---------I------I-----I------------------------------------------

    subroutine BANDOPTIM(graph, vertix, lmatbk, permut, &
         iverbose, iounit, myproc, nproc, tids, nmsg)

      ! Formal parameters
      integer, dimension (:) :: graph, vertix, permut
      integer, dimension (:) :: lmatbk, tids
      integer :: iverbose, iounit, myproc, nproc, nmsg
      intent (in) :: graph, vertix, lmatbk, iverbose, iounit
      intent (in) :: myproc, nproc, tids
      intent (inout) :: nmsg
      intent (out) :: permut

      ! Local variables for all steps
      integer, dimension(nproc+1) :: ptrmbk
      integer :: frtid, lfrtid, totid, ltotid, mids, midr, ierr

      integer :: i, j, p, itmp, nodecount
      integer :: l, ltot, maxdeg
      integer :: cand
      integer :: u, uwidth, v, vdepth, vwidth, sdepth, smaxdepth
      integer :: smaxdidx, swidth, sminwidth, sminwidx
      integer, allocatable, dimension(:) :: degree, levelv, levelu
      integer, allocatable, dimension(:) :: levels, levell
      integer :: minmaxfield(3), minmaxout(3), minmaxwork(3)
      integer :: newnr, subparts, processed, unprocessed
      ! Local variables needed only by step 1
      integer :: myminidx, deg
      integer, allocatable, dimension(:,:) :: cands, ibuf11, ibuf12
      ! Local variables needed only by step 2
      integer :: c, pos, l0, h0
      integer, allocatable, dimension(:) :: ni, hi, li
      integer :: ccnt
      integer, allocatable, dimension(:) :: ccard, ibuf1, ibuf2
      integer :: index_for_c1
      integer :: conflict(nproc)
      integer :: adjmat(nproc, nproc), adjbuf1(nproc, nproc)
      integer :: adjbuf2(nproc, nproc)
      ! Local variables needed only by step 3
      integer :: vglob
      logical :: step_a_xchanged
      logical, allocatable, dimension(:) :: mask
      integer :: lownewnr
      integer :: clevel, plevel, remaining
      integer, allocatable, dimension(:,:) :: clevnodes, clnbuf1, clnbuf2
      integer :: candnewnum, newnums
      integer, allocatable, dimension(:) :: nodes, degs, sortorder
      integer, allocatable, dimension(:) :: newnodes
      logical :: found

!!VAMPITRACE calls left out

      ! Initialize parameters for the ringshift
      lfrtid=LL9MAP(myproc-1,nproc)
      frtid=tids(lfrtid)
      ltotid=LL9MAP(myproc+1, nproc)
      totid=tids(ltotid)

      ! Initialize ptrmbk
      ptrmbk(1)=0
      do p=1, nproc
        ptrmbk(p+1)=ptrmbk(p)+lmatbk(p)
      enddo

      ! Determine local and global dimension of the matrix
      l=maxval(lmatbk)
      ltot=ptrmbk(nproc+1)

      allocate(degree(l))

      ! Determine degree of vertices
      do i=1,lmatbk(myproc)
        degree(i)=vertix(i+1)-vertix(i)
      enddo

      ! Determine global maximum degree
      minmaxfield=maxval(degree(1:lmatbk(myproc)))
      call LL4INM(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
           myproc, nproc, tids, nmsg)
      maxdeg=minmaxout(1)

      ! To ensure that the finding of a minimal integer including
      ! resolving ties for lower processor numbers this must hold:
      ! max(degree)*(nproc+1)<huge(1)
      if ((real(maxdeg)*(nproc+1)) > huge(1)) then
        write(iounit,*) 'ERROR in BANDOPTIM: max(degree)*(nproc+1)>=huge(1)'
        stop
      endif

      ! The allocates and deallocates fit pairwise
      allocate(levell(l))

      subparts=1
      permut=0
      newnr=1

      ! Pre-processing phase:
      !   Single main diagonal entries have to be permuted to the front
1     continue

      ! Note that we have to jump back here so we reallocate
      ! all necessary arrays
      allocate(cands(nproc,maxdeg), ibuf11(nproc,maxdeg), &
           ibuf12(nproc,maxdeg))
      cands(1:nproc,1)=0
      main_diag: if (subparts==1) then
        do i=1, lmatbk(myproc)
          if (degree(i) == 0) then
            cands(myproc,1)=cands(myproc,1)+1
          endif
        enddo
        call LL4ISM(1, nproc, 1, cands, ibuf11, ibuf12, &
             myproc, nproc, tids, nmsg)
        cands(1:nproc,1)=ibuf11(1:nproc,1)
        do p=1, nproc
          if ((p == myproc) .and. (cands(p,1)/=0)) then
            do i=1, lmatbk(myproc)
              if (degree(i) == 0) then
                permut(i)=newnr
                newnr=newnr+1
              endif
            enddo
          else
            newnr=newnr+cands(p,1)
          endif
        enddo
        if (myproc==1) then
          if ((newnr>1).and.(iverbose>0)) then
            write (iounit,*) 'Warning - ',newnr-1,' main diagonal entries'
          endif
        endif
      endif main_diag

      ! Phase I:
      !   Determine a pseudo diameter (from v to u)

      ! (Step G2) Find a local minum degree node and globalize it.
      !           Ties are broken by favouring the lower numbered proc

!!VAMPITRACE calls left out
      allocate(levelu(l), levelv(l))
      allocate(levels(l))

      minmaxfield(1)=maxdeg+1
      do i=1, lmatbk(myproc)
        if ((permut(i)==0) .and. (degree(i) < minmaxfield(1))) then
          minmaxfield(1)=degree(i)
          myminidx=i
        endif
      enddo
      minmaxfield(1)=minmaxfield(1)*nproc+(myproc-1)
      call LL4MIN(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
           myproc, nproc, tids, nmsg)
      if ((minmaxfield(1) == minmaxout(1))) then
        v=myminidx
      else
        v=0
      endif

      ! (Step G3)
      levelv=0
      call setlevel(graph, vertix, levelv, v, 1, vwidth, vdepth)
      newv: do
        cands(1:nproc,1:maxdeg)=0
        sminwidth=ltot
        ! On multiple processors synchronization is achieved by
        ! predetermining how many (reverse) level-structures are
        ! calculated per processor ordered by degrees.
        ! This way we can circumvent a sorting over the processors
        if (nproc>1) then
          do i=1, lmatbk(myproc)
            if (levelv(i)==vdepth) then
              cands(myproc,degree(i))=cands(myproc,degree(i))+1
            endif
          enddo
          call LL4ISM(1, nproc*maxdeg, 1, cands, ibuf11, ibuf12, &
               myproc, nproc, tids, nmsg)
          cands=ibuf11
        else
          cands(1,1)=lmatbk(myproc) ! Force to loop over all local els
        endif

        ! (Step G4)
        do deg=1, maxdeg
          do p=1,nproc
            cand=0
            per_proc: do i=1,cands(p,deg)
              if (p==myproc) then
                ! Search next candidate
                do j=cand+1, lmatbk(myproc)
                  if (levelv(j)==vdepth) then
                    cand=j
                    exit
                  endif
                enddo
                if (j>lmatbk(myproc)) exit per_proc
              endif
              levels=0
              call setlevel(graph, vertix, levels, cand, 1, swidth, sdepth)
              ! When we found a vertex s with a deeper level structure then
              ! set v <- s and reiterate
              if (sdepth > vdepth) then
                ! Save level structure for next loop
                v=cand
                levelv=levels
                vwidth=swidth
                vdepth=sdepth
                cycle newv
              endif

              if (swidth < sminwidth) then
                sminwidth=swidth
                sminwidx=cand
                levelu=levels
              endif
            enddo per_proc
          enddo
        enddo

        ! If we get here we can stop
        exit newv
      enddo newv

      deallocate(cands, ibuf11, ibuf12)

      u=sminwidx
      uwidth=sminwidth
      ! Now levelv, v, vwidth, levelu, u and uwidth are known

      ! Phase II:
      !   Combine the level structures of u and v to form a new level
      !   structure l with a smaller width

!!VAMPITRACE calls left out

      ! Step G6:
      ! Determine the associated level pairs (levelv, levelu)
      ! Also mark entries that have already been numbered or
      ! cannot be reached from the current starting node
      minmaxfield(1)=count(levelu/=0)
      minmaxfield(2)=count((permut==0).and.(levelu==0)) &
           -(l-lmatbk(myproc))
      call LL4ISM(1, 2, 1, minmaxfield, minmaxout, minmaxwork, &
           myproc, nproc, tids, nmsg)
      processed=minmaxout(1)
      unprocessed=minmaxout(2)

      do i=1, lmatbk(myproc)
        if ((permut(i)/=0) .or. (levelu(i)==0))  then
          ! Mask these entries out
          levelu(i)=vdepth+1
          levelv(i)=vdepth+1
        else
          levelu(i)=vdepth+1-levelu(i)
        endif
      enddo

      deallocate(levels)
      allocate(ni(vdepth))

      ! Assign all vertices with (i, i) - level pairs to level structure l
      ! (Step G7)
      where (levelu==levelv)
        levell=levelu
      elsewhere
        levell=0
      end where
      ! Compute ni upfront (step G9-1)
      do i=1, vdepth
        ni(i)=count(levell==i)
      enddo

      index_for_c1=0
      ! Find all connected components
      ! (Step G8)
      find_ccs: do c=myproc,l*nproc,nproc
        pos=0
        do i=1, lmatbk(myproc)
          if (levell(i)==0) then
            pos=i
            exit
          endif
        enddo

        ! Are there any more components to process?
        minmaxfield(1)=pos
        call LL4INM(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
             myproc, nproc, tids, nmsg)
        if (minmaxout(1) == 0) exit find_ccs

        ! Determine connected component with setlevel
        call setlevel(graph, vertix, levell, pos, -c, swidth, sdepth)
      enddo find_ccs

      ! Now calculate the cardinality of all components and iterate over
      ! them in order of decreasing cardinality
      ccnt=((c-1)/nproc)*nproc
      allocate(ccard(ccnt))
      ccard=0

      do i=1, lmatbk(myproc)
        if (levell(i)<0) then
          ccard(-levell(i))=ccard(-levell(i))+1
        endif
      enddo
      if (nproc>1) then
        allocate(ibuf1(ccnt),ibuf2(ccnt))
        call LL4ISM(1, ccnt, 1, ccard, ibuf1, ibuf2, &
             myproc, nproc, tids, nmsg)
        ccard=ibuf1
        deallocate(ibuf1,ibuf2)
      endif

      allocate(hi(vdepth), li(vdepth), ibuf1(vdepth), ibuf2(vdepth))

      if (nproc>1) then
        ! Globalize ni
        call LL4ISM(1, vdepth, 1, ni, ibuf1, ibuf2, &
             myproc, nproc, tids, nmsg)
        ni=ibuf1
      endif

      ! Assign the connected components
      ! (Step  G9)
      assign_c: do c=1, c+1
        pos=0
        minmaxfield(1)=0
        do i=1, ccnt
          if (ccard(i)>minmaxfield(1)) then
            minmaxfield(1)=ccard(i)
            pos=i
          endif
        enddo

        if (pos==0) exit assign_c
        ccard(pos)=0

        ! (Step G9-2)
        hi=0
        li=0
        do i=1, lmatbk(myproc)
          if (levell(i)==(-pos)) then
            hi(levelv(i))=hi(levelv(i))+1
            li(levelu(i))=li(levelu(i))+1
          endif
        enddo

        if (nproc>1) then
          ! Globalize hi and li
          call LL4ISM(1, vdepth, 1, hi, ibuf1, ibuf2, &
               myproc, nproc, tids, nmsg)
          hi=ibuf1
          call LL4ISM(1, vdepth, 1, li, ibuf1, ibuf2, &
               myproc, nproc, tids, nmsg)
          li=ibuf1
        endif

        hi=hi+ni
        li=li+ni

        ! Depending on which assignment yields the narrower level
        ! structure assign all components as indicated by either
        ! levelv or leveluu
        ! (Steps G9-3 to G9-6)
        h0=maxval(hi, 1, hi-ni>0)
        l0=maxval(li, 1, li-ni>0)
        if ((h0<l0) .or. ((h0==l0) .and. (vwidth<=uwidth))) then
          if (c==1) index_for_c1=1
          where (levell==(-pos))
            levell=levelv
          end where
          ni=hi
        else
          if (c==1) index_for_c1=2
          where (levell==(-pos))
            levell=levelu
          end where
          ni=li
        endif
      enddo assign_c

      deallocate(hi, li, ibuf1, ibuf2)
      deallocate(ccard)
      deallocate(ni)
      deallocate(levelu, levelv)

      ! Phase III:
      !   (Re-)Number the vertices

!!VAMPITRACE calls left out

      allocate(nodes(maxdeg), degs(maxdeg), sortorder(maxdeg))
      allocate(newnodes(maxdeg))
      if (nproc>1) then
        allocate(clevnodes(vwidth,3), clnbuf1(vwidth,3), clnbuf2(vwidth,3))
        allocate(ibuf1(maxdeg), ibuf2(maxdeg))
      endif

      ! (Step G10)
      ! Broadcast degree(u) and degree(v) and if
      ! degree(u)<degree(v) then swap u and v and reverse the
      ! level structure
      minmaxfield=0
      if (u/=0) minmaxfield(1)=degree(u)
      if (v/=0) then
        minmaxfield(2)=degree(v)
        minmaxfield(3)=globalize(v)
      endif
      call LL4ISM(1, 3, 1, minmaxfield, minmaxout, minmaxwork, &
           myproc, nproc, tids, nmsg)
      minmaxfield=minmaxout
      vglob=minmaxout(3)
      if (minmaxfield(1)<minmaxfield(2)) then
        ! Exchange u and v
        itmp=v; v=u; u=itmp
        levell=vdepth+1-levell
        ! Keep the markings on the processed nodes
        where (levell==0) levell=vdepth+1
        step_a_xchanged=.true.
      else
        step_a_xchanged=.false.
      endif

      ! (Step G11/12-1)
      clevel=1
      plevel=1
      ! We can fill this in now on the processor owning v, we have to
      ! do this on the other procs when clevnodes() has been filled
      lownewnr=newnr
      if (v/=0) permut(v)=newnr
      newnr=newnr+1
      for_all_levels: do clevel=1, vdepth
        if (nproc>1) then
          ! Gather information about nodes in the level clevel
          ! by a ring 
          nodecount=0
          remaining=0
          clevnodes(:,1:2)=0
          clnbuf2=0
          do i=1,lmatbk(myproc)
            if (levell(i)==clevel) then
              nodecount=nodecount+1
              clnbuf2(nodecount,1)=globalize(i)
              clnbuf2(nodecount,2)=degree(i)
            endif
            if ((levell(i)==clevel) .and. (permut(i)==0)) &
                 remaining=remaining+1
          enddo

          minmaxfield(1)=remaining
          call LL4ISM(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
               myproc, nproc, tids, nmsg)
          remaining=minmaxout(1)

          nodecount=0   ! Our local field is empty right now
          p=1
          call MPRCVA(frtid, nmsg+p, 2*vwidth*iint, clnbuf1, midr, ierr)
          call MPSNDA(totid, nmsg+p, 2*vwidth*iint, clnbuf2, mids, ierr)
          call MPSNDW(totid, nmsg+p, 2*vwidth*iint, clnbuf2, mids, ierr)
          call MPRCVW(frtid, nmsg+p, 2*vwidth*iint, clnbuf1, midr, ierr)
          ringshift: do p=2, nproc
            clnbuf2=clnbuf1
            call MPRCVA(frtid, nmsg+p, 2*vwidth*iint, clnbuf1, midr, ierr)
            call MPSNDA(totid, nmsg+p, 2*vwidth*iint, clnbuf2, mids, ierr)

            ! Sort the node information that is passing us by into
            ! the local table
            j=1
            do i=1, vwidth
              if (clnbuf2(i,1)==0) exit
              do while ((j<=nodecount) .and. (clevnodes(j,1)<clnbuf2(i,1)))
                j=j+1
              enddo
              do itmp=nodecount,j,-1
                clevnodes(itmp+1,1:2)=clevnodes(itmp,1:2)
              enddo
              clevnodes(j,1:2)=clnbuf2(i,1:2)
              nodecount=nodecount+1
            enddo

            call MPSNDW(totid, nmsg+p, 2*vwidth*iint, clnbuf2, mids, ierr)
            call MPRCVW(frtid, nmsg+p, 2*vwidth*iint, clnbuf1, midr, ierr)
          enddo ringshift
          nmsg=nmsg+nproc
          ! Blank out the (non-transmitted) buffer area for the new
          ! numbers
          clevnodes(:,3)=0
        else
          ! If we are running serially then we only have to
          ! determine how many candidates to number in this step
          remaining=0
          do i=1, lmatbk(myproc)
            if ((levell(i)==clevel) .and. (permut(i)==0)) then
              remaining=remaining+1
            endif
          enddo
        endif

        ! Now we can fill in the first node on all other
        ! processors
        if ((v==0) .and. (clevel==1)) then
          pos=findidx(vglob, clevnodes, nodecount)
          clevnodes(pos, 3)=lownewnr
        endif

        do while (remaining > 0)
          ! (Step G11/12-2)
          ! Find lowest numbered vertex in level plevel that has
          ! unnumbered neighbours in clevel
          cand=ltot+1
          candnewnum=ltot+1
          search: do i=1,lmatbk(myproc)
            if ((levell(i)/=plevel) .or. (permut(i)==0)) cycle search
            do j=1, degree(i)
              if (islocaltop(graph(vertix(i)+j), myproc)) then
                ! The neighbour is on our processor
                if ((levell(localize(graph(vertix(i)+j)))==clevel) &
                     .and. (permut(localize(graph(vertix(i)+j)))==0)) then
                  if ((cand==ltot+1) .or. (permut(i)<candnewnum)) then
                    cand=globalize(i)
                    candnewnum=permut(i)
                  endif
                endif
              else
                ! Look up the neighbour in the buffer
                pos=findidx(graph(vertix(i)+j), clevnodes, nodecount)
                if ((pos/=0) .and. (clevnodes(pos,3)==0)) then
                  if ((cand==ltot+1) .or. (permut(i)<candnewnum)) then
                    cand=globalize(i)
                    candnewnum=permut(i)
                  endif
                  cycle search
                endif
              endif
            enddo
          enddo search

          ! The processor posessing the candidate with the lowest
          ! number assigns the next numbers
          minmaxfield(1)=candnewnum
          call LL4MIN(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
               myproc, nproc, tids, nmsg)
          candnewnum=minmaxout(1)
          if (minmaxout(1)/=ltot+1) then
            if ((minmaxfield(1) == minmaxout(1))) then
              ! This processor assigns the next numbers
              newnums=0
              cand=localize(cand)
              do j=1,degree(cand)
                if (islocaltop(graph(vertix(cand)+j), myproc)) then
                  pos=localize(graph(vertix(cand)+j))
                  if ((levell(pos)==clevel) .and. &
                       (permut(pos)==0)) then
                    newnums=newnums+1
                    nodes(newnums)=globalize(pos)
                    degs(newnums)=degree(pos)
                  endif
                else
                  pos=findidx(graph(vertix(cand)+j), &
                       clevnodes, nodecount)
                  if ((pos/=0) .and. (clevnodes(pos,3)==0)) then
                    newnums=newnums+1
                    nodes(newnums)=clevnodes(pos,1)
                    degs(newnums)=clevnodes(pos,2)
                  endif
                endif
              enddo
              ! Sort the nodes by their degree (but do not generate
              ! overhead for the probable case of newnums==1)
              if (newnums>1) then
                call HEAPSORTIX(degs, sortorder, newnums, .true.)
              else
                sortorder(1)=1
              endif
              newnodes=0
              do j=1, newnums
                if (islocaltop(nodes(sortorder(j)), myproc)) then
                  permut(localize(nodes(sortorder(j))))=newnr
                endif
                newnodes(j)=nodes(sortorder(j))
                newnr=newnr+1
              enddo
              remaining=remaining-newnums
              ! Broadcast the nodes that were numbered (in the correct
              ! order) -> This statement matches the LL4ISM call below
              if (nproc>1) then
                call LL4ISM(1, maxdeg, 1, newnodes, ibuf1, ibuf2, &
                     myproc, nproc, tids, nmsg)
              endif
              ! Now we can store non-local new numbers in our buffer
              do i=1,maxdeg
                if (newnodes(i)==0) exit
                if (.not. islocaltop(newnodes(i), myproc)) then
                  pos=findidx(newnodes(i), clevnodes, nodecount)
                  clevnodes(pos,3)=newnr-(newnums-i)-1
                endif
              enddo
            else
              ! Receive the new numbers and store them in the
              ! local table (note that we pass through here only
              ! for nproc>1)
              newnodes=0
              call LL4ISM(1, maxdeg, 1, newnodes, ibuf1, ibuf2, &
                   myproc, nproc, tids, nmsg)
              newnodes=ibuf1
              do i=1,maxdeg
                if (newnodes(i)==0) exit
                if (.not. islocaltop(newnodes(i), myproc)) then
                  pos=findidx(newnodes(i), clevnodes, nodecount)
                  clevnodes(pos,3)=newnr
                else
                  permut(localize(newnodes(i)))=newnr
                endif
                newnr=newnr+1
                remaining=remaining-1
              enddo
            endif
          else
            ! (Step G11/12-3)
            ! If we still have unnumbered vertices, then number the lowest
            cand=ltot+1
            do i=1,lmatbk(myproc)
              if ((levell(i)==clevel) .and. (permut(i)==0)) then
                cand=globalize(i)
                exit
              endif
            enddo
            minmaxfield(1)=cand
            call LL4MIN(1, 1, 1, minmaxfield, minmaxout, minmaxwork, &
                 myproc, nproc, tids, nmsg)
            cand=minmaxout(1)
            if (minmaxfield(1)==minmaxout(1)) then
              permut(localize(cand))=newnr
            else
              pos=findidx(cand, clevnodes, nodecount)
              clevnodes(pos,3)=newnr
            endif
            newnr=newnr+1
            remaining=remaining-1
          endif
        enddo
        plevel=clevel
      enddo for_all_levels

      ! (Step III.D)
      if ((step_a_xchanged .and. (index_for_c1==2)) .or. &
          ((.not. step_a_xchanged) .and. (index_for_c1==1))) then
        where (levell/=vdepth+1)
          permut=processed-permut+2*lownewnr-1
        endwhere
      endif

      if (nproc>1) then
        deallocate(ibuf1, ibuf2)
        deallocate(clevnodes, clnbuf1, clnbuf2)
      endif
      deallocate(newnodes)
      deallocate(nodes, degs, sortorder)

      ! If we still have unprocessed nodes then do another
      ! round
      if (ptrmbk(nproc+1)/=(newnr-1)) then
        subparts=subparts+1
        goto 1
      endif

      if (subparts > 1) then
        if ((myproc==1).and.(iverbose>0)) then
          write (iounit,*) 'Warning - matrix can be decomposed in ', &
               subparts,' independent problems.'
        endif
      endif

      deallocate(levell)
      deallocate(degree)

!!VAMPITRACE calls left out

    contains


      ! Subroutine : setlevel
      ! Purpose    : Determine the level structure of (GRAPH, VERTIX)
      !              beginning with STARTNODE at STARTLEV.
      !              LEVEL is the calculated level structure having WIDTH and
      !              DEPTH.
      subroutine setlevel(graph, vertix, level, startnode, startlev, &
           width, depth)

        ! Formal parameters
        integer, dimension (:) ::  graph, vertix, level
        integer :: startnode, startlev, width, depth
        intent (in) graph, vertix, startnode
        intent (inout) level
        intent (out) width, depth

        ! Local variables
        integer :: i, j, k, p, nolev, tlev, l, nonb, cwidth, nrblk
        integer, allocatable, dimension(:) :: lold, ldiff, lbuf
        integer :: ldsum(1)
        integer :: iwork(nproc)
        integer :: confloff, c1, c2

!!VAMPITRACE calls left out
        nolev=startlev
        l=maxval(lmatbk)
        allocate(lold(l), ldiff(l), lbuf(l))
        adjmat=0

        ! Initialize conflict list for connected components
        if (startlev<0) then
          confloff=(((-startlev-1))/nproc) * (-nproc)
          do i=1,nproc
            conflict(i)=confloff - i
          enddo
        endif

        width=0
        lold=level

        ! If this processor does not posess the root node then do nothing
        ! in the first cycle
        if (startnode /= 0) then
          level(startnode)=nolev
        endif

        next_level: do
          cwidth=0
          ldiff=level-lold
          call LL4ISM(l, 1, l, ldiff, ldsum(1), iwork(1), &
               myproc, nproc, tids, nmsg)
          if (ldsum(1) == 0) exit next_level
          cwidth=ldsum(1)/nolev        ! See warning below
          if (cwidth > width) then
            width=cwidth
          end if
          ! Do not adjust the level for connectivity considerations
          if (nolev > 0) nolev=nolev+1
          lold=level

          ! Here comes the ringshift across all processors
          ! Note: There can be no latency hiding here because
          !       the static information (neighbourhood-relations)
          !       is bound to the processors. Otherwise the (non-fixed size)
          !       neighbour-relations have to be passed around. Could
          !       be worth checking.
          procs: do p=1, nproc
            if (nproc>1) then
              call MPRCVA(frtid, nmsg+p, l*iint, lbuf, midr, ierr)
            endif
            ! fast version of nrblk=(((myproc-p+1)-1) mod nproc)+1
            nrblk=myproc+nproc-p+1
            if (nrblk>nproc) nrblk=nrblk-nproc

            do i=1,lmatbk(myproc)
              if (ldiff(i)/=0) then
                if (ldiff(i)<0) then
                  tlev=ldiff(i)
                else
                  tlev=nolev
                endif

                nonb=vertix(i+1)-vertix(i)
                neighbours: do j=1,nonb
                  if (islocaltop(graph(vertix(i)+j), nrblk)) then
                    if (level(graph(vertix(i)+j)-ptrmbk(nrblk))==0) then
                      level(graph(vertix(i)+j)-ptrmbk(nrblk))=tlev
                    elseif (level(graph(vertix(i)+j)-ptrmbk(nrblk))<0 .and. &
                         level(graph(vertix(i)+j)-ptrmbk(nrblk))/=tlev) then
                      ! Note the conflict in the adjacency matrix
                      c1=-(tlev-confloff)
                      c2=-(level(graph(vertix(i)+j)-ptrmbk(nrblk))-confloff)
                      adjmat(c1,c2)=1
                      adjmat(c2,c1)=1
                    endif
                  endif
                enddo neighbours

              endif
            enddo
            if (nproc>1) then
              call MPSNDA(totid, nmsg+p, l*iint, level, mids, ierr)
              call MPSNDW(totid, nmsg+p, l*iint, level, mids, ierr)
              call MPRCVW(frtid, nmsg+p, l*iint, lbuf, midr, ierr)
              level=lbuf
            endif
          enddo procs
          if (nproc>1) nmsg=nmsg+nproc
        enddo next_level

        deallocate(lold, ldiff, lbuf)

!!VAMPITRACE calls left out

        ! If we are determining connected components on multiple
        ! processors then we have to globalize the local adjacency
        ! matrix through a global sum
        ccadjust: if ((startlev<0) .and. (nproc>1)) then
          call LL4ISM(1, nproc*nproc, 1, adjmat, adjbuf1, adjbuf2, &
               myproc, nproc, tids, nmsg)
          adjmat=adjbuf1

          ! Now we can mark all connected components
          conflict=0
          do i=1, nproc
            if (conflict(i)==0) then
              call mark(i,confloff-i)
            endif
          enddo

          ! And post-process our level information
          do i=1, lmatbk(myproc)
            if (level(i)<confloff) then
              level(i)=conflict(confloff-level(i))
            endif
          enddo
        endif ccadjust

        ! Warning: depth is only sensible when startlev==1
        ! Warning: width is only sensible when startlev is equal on all
        !          processors (i.e. not when considering cc's)
        !          Otherwise width will be zero
        depth=nolev-1

!!VAMPITRACE calls left out
        return
      end subroutine setlevel



      ! Subroutine : mark
      ! Purpose    : Mark the connected component beginning at START
      !              with NUM
      !              The adjacency list is in ADJMAT and the numbering
      !              is done in CONFLICT
      recursive subroutine mark(start, num)
        integer :: start, num
        intent(in) start, num
        integer i

        conflict(start)=num
        do i=1,nproc
          if (adjmat(start, i)/=0) then
            if (conflict(i)==0) then
              call mark(i,num)
            endif
          endif
        enddo
      end subroutine mark


      ! Function   : islocaltop
      ! Purpose    : Determine whether NODE is local to processor
      !              PROC using the information in (module-wide) PTRMBK
      logical function islocaltop(node, proc)
        ! Formal parameters
        integer :: node, proc
        intent (in) node, proc

        islocaltop=(node>ptrmbk(proc) .and. node<=ptrmbk(proc+1))
      end function islocaltop

      ! Function   : localize
      ! Purpose    : Determine the local node number for NODE
      integer function localize(node)
        ! Formal parameters
        integer :: node
        intent (in) node

        localize=node-ptrmbk(myproc)
      end function localize

      ! Function   : globalize
      ! Purpose    : Determine the global node number for NODE
      integer function globalize(node)
        ! Formal parameters
        integer :: node
        intent (in) node

        globalize=node+ptrmbk(myproc)
      end function globalize


      ! Function   : findidx
      ! Purpose    : Find the index of element EL in the sorted
      !              one dimensional array A of size N.
      integer function findidx(el, a, n)
        ! Formal parameters
        integer :: el, n
        integer, dimension (:) ::  a(n)
        intent (in) el, a, n

        ! Local variables
        integer :: idx, lower, upper
        logical :: found

        found=.false.
        lower=1
        upper=n
        do while (lower<=upper)
          idx=(lower+upper)/2
          if (a(idx)==el) then
            found=.true.
            exit
          endif
          if (el<a(idx)) then
            upper=idx-1
          else
            lower=idx+1
          endif
        enddo

        if (found) then
          findidx=idx
        else
          findidx=0
        endif

      end function findidx

    end subroutine BANDOPTIM

end module bando
